function [u_mid,err] = fast_poisson_with_estimate(rectangle,f,bdry,h);

% u_mid = est_fast_poisson(rectangle,f,bdry,h)
%
% solves poisson equation with dirichlet boundary conditions
% on rectangle [0,a] x [0,b] using a five-point stencil;
% solver: fast sine transform
%
%   rectangle   [a,b]
%               a   length in x-direction
%               b   length in y-direction
%   f           right-hand side of -\Delta u(x) = f (constant)
%   bdry        Dirichlet data [x0,xa,y0,yb]
%               x0  boundary value on {0} x [0,b]   (constant)
%               xa  boundary value on {a} x [0,b]   (constant)
%               y0  boundary value on [0,a] x {0}   (constant)
%               yb  boundary value on [0,a] x {b}   (constant)
%   h   h = 1/n suggested approximate grid-size
%   u_mid   solution at the midpoint
%   err     estimated relative error of u_mid
%
% algorithm as in J. Demmel, Applied Numerical Linear Algebra,
% SIAM 1997, Section 6.7
%
% written by F. Bornemann, 20.2.2003

% the grid
n = 2*ceil(1/2/h);                              % make n even
a = rectangle(1); b = rectangle(2);
nx = ceil(a)*n-1; ny = ceil(b)*n-1; x=1:nx; y=1:ny; [nx,ny]
dx = a/(nx+1); dy = b/(ny+1);

% the right hand side
b = f*ones(nx,ny); 
b(1,y) = b(1,y)+bdry(1)/dx^2; b(nx,y) = b(nx,y)+bdry(2)/dx^2;
b(x,1) = b(x,1)+bdry(3)/dy^2; b(x,ny) = b(x,ny)+bdry(4)/dy^2;

% solve the problem by 2D fast sine transform
u = dst2(b);    
d = 4*(sin(x'/2*pi/(nx+1)).^2*ones(1,ny)/dx^2+ ...
                        ones(nx,1)*sin(y/2*pi/(ny+1)).^2/dy^2);
u = u./d;       
u = dst2(u);   

% the matrix
Ax = spdiags(ones(nx,1)*[-1 2 -1]/dx^2,-1:1,nx,nx);
Ay = spdiags(ones(ny,1)*[-1 2 -1]/dy^2,-1:1,ny,ny);
A  = kron(Ay,speye(nx)) + kron(speye(ny),Ax);

% estimate relative error of solution components
% see N. Higham, Accuracy and Stability of Numerical Algorithms, 
% SIAM 1996, 7.7, 14.1, 14.3
r = b(:) - A*u(:); 
% ga = sqrt((nx*ny+1))*eps/2; ga = ga/(1-ga);
r = abs(r); % + ga*(abs(A)*abs(u(:))+abs(b(:)));
D1 = spdiags(1./abs(u(:)),0,nx*ny,nx*ny);
D2 = spdiags(r,0,nx*ny,nx*ny);
err = normest1(@afun,[],[],nx,ny,dy,dx,D1,D2);

% the solution
u_mid = u((nx+1)/2,(ny+1)/2);

return

% =============================================================
% routines for 1D and 2D fast sine transform, cf. Demmel p. 324
% =============================================================

function y = dst(x) 
n = size(x,1); m = size(x,2);
y = [zeros(1,m);x]; y = imag(fft(y,2*n+2)); y = sqrt(2/(n+1))*y(2:n+1,:);
return

function y = dst2(x)
y = dst(dst(x)')';
return

% afun routine for Matlab's normest1 command

function result = afun(flag,b,nx,ny,dy,dx,D1,D2)

switch flag
    case 'dim'
        result = nx*ny;
    case 'real'
        result = 1;
    case 'notransp'
        x = 1:nx; y=1:ny;
        [n,m] = size(b);
        for j=1:m
            r = reshape(D1*b(:,j),nx,ny);
            u = dst2(r);    
            d = 4*(sin(x'/2*pi/(nx+1)).^2*ones(1,ny)/dx^2+ ...
                        ones(nx,1)*sin(y/2*pi/(ny+1)).^2/dy^2);
            u = u./d;       
            u = dst2(u);   
            result(:,j) = D2*u(:);
        end
    case 'transp'
        x = 1:nx; y=1:ny;
        [n,m] = size(b);
        for j=1:m
            r = reshape(D2*b(:,j),nx,ny);
            u = dst2(r);    
            d = 4*(sin(x'/2*pi/(nx+1)).^2*ones(1,ny)/dx^2+ ...
                        ones(nx,1)*sin(y/2*pi/(ny+1)).^2/dy^2);
            u = u./d;       
            u = dst2(u);   
            result(:,j) = D1*u(:);
        end
end

return;